set apsTmpStringCount 0
proc APSTmpString { } {
    global env apsTmpStringCount
    set tString [clock format [clock seconds] -format %y%m%d-%H%M%S]-[pid]
    if {[info exists env(USER)]} {
        set tString ${tString}$env(USER)$apsTmpStringCount
    } else {
        set tString ${tString}$apsTmpStringCount
    }
    incr apsTmpStringCount
    regsub -all " " $tString "" tString1
    return $tString1
}
proc APSParseArguments {optlist} {
    upvar args arguments
    set length [llength $arguments]
    set index 0
    set leftovers {}
    while {$index<$length} {
        set arg [lindex $arguments $index]
        if {[string index $arg 0]=="-"} {
            set keywordName [string range $arg 1 end]
            if {[lsearch -exact $optlist $keywordName]!=-1} {
                incr index
                if {$index==$length} {
                    lappend leftovers $arg
                } else {
                    set valueString [lindex $arguments $index]
                    uplevel "set $keywordName {$valueString}"
                    incr index
                }
            } else {
                incr index
                lappend leftovers $arg
            }
        } else {
            lappend leftovers $arg
            incr index
        }
    }
    set arguments [concat $leftovers]
}
#-------------------------------------------------------------------------------------------------------------------------
proc Fit_MakeElementNameFromTagName {args} {
    #------ elegant can create a mixture of names when some names have #N in it and others don't.
    #------ To handle it, add #1 at the end, then delete second # symbol and after it.
    #------ Also delete #X and #Y
    set output ""
    set tagNameColumn TagName
    set elementNameColumn ElementName
    set addElementOccurence 0
    APSParseArguments {input output tagNameColumn elementNameColumn addElementOccurence}
    set columnList [join [exec sddsquery $input -col] ]
    set localFile $input.tmp
    if {[lsearch -exact $columnList $elementNameColumn] == -1} {
	exec sddsprocess $input $input.tmp "-edit=col,$elementNameColumn,$tagNameColumn,@#X@@ @#Y@@ S?@#@ K"
    } else {
	file copy -force $input $input.tmp
    }
    if $addElementOccurence {
	#--- Overwrite if exists:
	exec sddsconvert $input.tmp -pipe=out -del=col,ElementOccurence \
	    | sddsprocess -pipe "-edit=col,TagNameCopy,$tagNameColumn,@#X@@ @#Y@@ e i@#1@ a 2S?@#@ 2d" \
	    "-scan=col,ElementOccurence,TagNameCopy,%ld,type=long,edit=Z#" \
	    | sddsconvert -pipe=in $input.tmp1 -del=col,TagNameCopy
	set localFile $input.tmp1
    }
    if [string length $output] {
	file copy -force $localFile $output
    } else {
	file copy -force $localFile $input
    }
    catch {file delete $input.tmp $input.tmp1}
}
#-------------------------------------------------------------------------------------------------------------------------
#------ This procedure reads variables from definitionFile and makes them global
proc Fit_ReadDefinitionFile {args} {
    set variableList ""
    set warnings 1
    set abortOnError 1
    APSParseArguments {definitionFile workDir variableList warnings abortOnError}

    #------ List of required variables in the definition file:
    #------ List1 is definition of internally used files:
    set requiredList1 [list dispColumn rmDerivFile rmDerivInverseFile rmDerivFileS rmDerivFileV rmDerivFileSV \
			   rmDerivFileU rmDifferenceFile dispersionFile hrmMeasured vrmMeasured hrmDifference \
			   vrmDifference hrmIterFile vrmIterFile elementsUpdateFile noCorrectorsFile weightFile \
			   manualWeightFile iterationsLogFile weightCorrectorsFile variableFile allElementsFile \
			   averageErrorFile nonlinFile elementOrderFile dispMeasuredFile rmDerivElementFile \
			   sigmaWeightFileX sigmaWeightFileY specialElementsInputFile specialParamFile variableArchiveFile \
			   iterationsLogFileSDDS]
    #------ List2 is various variables defined by user that are too many to input from the GUI:
    set requiredList2 [list hCalibFactor vCalibFactor dispCalibFactor dispFitWeight thetaKickX thetaKickY \
			   bRho zeroBpmTol kickRMD matlabScriptDir directRMCalc fixedOrbitLength \
			   offDiagWeight adjustAverageGains optimNonlin locoSearchString \
			   showMatchingMachineConfigOnly twiFile svdProgramName closedOrbit useDoubleOrbits \
			   useQsubLong useQsubShort numberOfQsubTasksLong numberOfQsubTasksShort waitTimeLong \
			   waitTimeShort waitIntervalLong waitIntervalShort qsubCommandLong qsubCommandShort \
			   queueSystemNameShort queueSystemNameLong qsubRespProcCommandShort qsubRespProcCommandLong \
			   submissionPauseLong submissionPauseShort useQuadConstraints badBpmSigmaLimit \
			   rmdQuadDelta rmdTiltDelta tiltCorFraction remotePath verbose useSigmaWeightFile \
			   simplifiedTiltTreatment]
    set requiredList [concat $requiredList1 $requiredList2]
    #------ This list is for variables that are not required but possible:
    set possibleList [list twissRefFile twissRefElement malignElement useIntercept useSlope useCubic dontUsePrevQuadConstr \
			  convergenceTolerance convergenceTolerance0 convergenceTolerance1 convergenceTolerance2 resErrorMax \
			  useParallelMultiplication matrixDivisions badBpmNSigmasPlotting elemByElemSR]

    #------ Reading only requested variables:
    if [string length $variableList] {
	if [catch {open $definitionFile r} fid] {
	    return -code error $fid
	}
	while {([gets $fid line] >= 0) && ([llength $variableList]> 0)} {
	    if {[string compare "\#" [string index $line 0]] != 0} {
		if {[string compare [lindex $line 0] "set"] == 0} {
		    set varName [lindex $line 1]
		    set index [lsearch -exact $variableList $varName]
		    if {$index != -1} {
			set varValue [lindex $line 2]
			uplevel "set $varName $varValue"
		    }
		    set variableList [lreplace $variableList $index $index]
		}
	    }
	}
	close $fid
	if [llength $variableList] {
	    if $abortOnError {
		return -code error "Error: the following variables were not defined in $definitionFile: $variableList"
	    } else {
		puts stdout "Warning: the following variables were not defined in $definitionFile: $variableList"
	    }
	}
	return
    }

    #------ Reading the entire definition file...
    if [catch {open $definitionFile r} fid] {
	return -code error $fid
    }
    set definedList ""
    while {[gets $fid line] >= 0} {
	if {[string compare "\#" [string index $line 0]] != 0} {
	    if {[string compare [lindex $line 0] "set"] == 0} {
		global [lindex $line 1]
		lappend definedList [lindex $line 1]
	    }
	}
    }
    close $fid
    source $definitionFile
    if $useQuadConstraints {lappend requiredList quadConstraintWeight quadConstraintFile}
    if !$closedOrbit {lappend requiredList twissRefFile twissRefElement malignElement}

    #------ Looking for missing and unnecessary variables...
    set unnecList ""
    foreach var $definedList {
	set ix1 [lsearch -exact $requiredList $var]
	set ix2 [lsearch -exact $possibleList $var]
	if {$ix1 != -1} {
	    set requiredList [lreplace $requiredList $ix1 $ix1]
	}
	if {$ix1 == -1 && $ix2 == -1} {
	    lappend unnecList $var
	}
    }
    if {[llength $unnecList] && $warnings} {
	OutputStatusMessage "Warning: following variables from $definitionFile are not required: $unnecList"
    }
    if [llength $requiredList] {
	return -code error "Error: following variables are required but are not in $definitionFile: $requiredList"
    }
}

#-------------------------------------------------------------------------------------------------------------------------
# This procedure makes BPM names unique by splitting MONI type into two elements HMON and 
# VMON and adding X or Y at the end of the name for each element. 
proc Fit_MakeBpmNamesUnique {args} {
    APSParseArguments {inputLTE outputLTE}

    if [catch {open $inputLTE r} fileId] {
	return -code error "Cannot open the file $inputLTE for reading: $fileId"
    }
    if [catch {open $outputLTE w} fileId1] {
	return -code error "Cannot open the file $outputLTE for writing: $fileId1"
    }
    while {[gets $fileId line] >= 0} {
	if {[string first "MONI" $line] != -1} {
	    set line [RemoveAllSpaces $line]
	    if {[string first "LINE=" $line] == -1} {
		set strLength [string length $line]
		if {[string compare "\"" [string range $line 0 0]] == 0} {
		    set I1 [expr 1 + [string first "\"" [string range $line 1 $strLength]]]
		    set colIndex [expr $I1 + [string first ":" [string range $line $I1 $strLength]]]
		    set elementName [string range $line 1 [expr $I1 - 1]]
		} else {
		    set colIndex [string first ":" $line]
		    set elementName [string range $line 0 [expr $colIndex - 1]]
		}
		if {[string first "MONI" [string range $line $colIndex $strLength]] != -1} {
		    puts $fileId1 "\"${elementName}X\": HMON"
		    puts $fileId1 "\"${elementName}Y\": VMON"
		    puts $fileId1 "\"${elementName}\":  LINE=(${elementName}X,${elementName}Y)"
		}
	    } else {
		puts $fileId1 $line
	    }
	} else {
	    puts $fileId1 $line
	}
    }
    close $fileId
    close $fileId1
}
#-------------------------------------------------------------------------------------------------------------------------
# This procedure creates list of bpms which have rms reading of less than the threshold.
proc Fit_GetZeroBpms { args } {
    set deleteFiles ""
    set abovebelow below
    APSParseArguments {filename threshold column abovebelow dispColumn}
    if {[lsearch -exact [join [exec sddsquery -col $filename] ] $column] == -1} {
	set filename1 [APSTmpString]-zeros
	exec sddstranspose $filename $filename1 -oldColumnNames=$column
	set filename $filename1
	lappend deleteFiles $filename1
    }
    switch -exact -- $abovebelow {
	below {set zeroBpms [join [exec sddsconvert $filename -pipe=out -del=col,$dispColumn \
				       | sddsrowstats -pipe -rms=Rms,* \
				       | sddsprocess -pipe -proc=Rms,rms,RmsRms \
				       | sddsprocess -nowarning -pipe "-test=col,Rms RmsRms $threshold * <" \
				       | sdds2stream -pipe=in -col=$column] ]
	}
	above {
	    if [catch {exec sddsconvert $filename -pipe=out -del=col,$dispColumn \
			   | sddsrowstats -pipe -rms=Rms,* \
			   | sddsoutlier -pipe -col=Rms -stdev=$threshold -passes=3 -inv \
			   | sdds2stream -pipe=in -col=$column} result] {
		set zeroBpms ""
	    } else {
		set zeroBpms [join $result ]
	    }
	}
	default {
	    return -code error "Wrong abovebelow argument: $abovebelow"
	}
    }
    if [llength $deleteFiles] {eval file delete $deleteFiles}
    return $zeroBpms
}

#-------------------------------------------------------------------------------------------------------------------------
# This procedure makes one-column file from a response matrix file (keeps 2 pages if matrixFile is 2-page).
proc Fit_MakeColumnFromMatrix {args} {
    set rootnameColumn Rootname
    APSParseArguments {matrixFile columnFile columnName rootnameColumn rowNameColumn}
    exec sddsmatrix2column $matrixFile $columnFile -rowNameColumn=$rowNameColumn -rootnameColumn=$rootnameColumn -dataColumn=$columnName -major=row
}
#-------------------------------------------------------------------------------------------------------------------------
# This procedure splits list into several lists. Useful for queuing elegant tasks.

proc Fit_SplitList { args } {
    
    APSParseArguments { List splitTasks }
    
    if {$splitTasks == 0} {
	return $List
    }
    set listLength [llength $List]
    if {$listLength < $splitTasks} {
	set split $listLength
    } else {
	set split $splitTasks
    }
    set N [expr $listLength / $split]
    set reminder [expr $listLength - $N \* $split]
    if {$reminder > $N} {
	while {$reminder > $N} {
	    incr split -1
	    set N [expr $listLength / $split]
	    set reminder [expr $listLength - $N \* $split]
	}
    }
    set ListOut ""
    for { set J 0 } { $J < $split } { incr J } {
	set L1 [expr $J \* $N]
	set L2 [expr \( $J + 1 \) \* $N]
	lappend ListOut [lrange $List $L1 [expr $L2 - 1]]
    }
    if {$reminder != 0} {
	if {$reminder < [expr $N / 4]} {
	    #--- If reminder is very small, append the rest to the last list from above (first delete the last list, then append):
	    set ListOut [lreplace $ListOut [expr $split - 1] [expr $split - 1]]
	    lappend ListOut [lrange $List $L1 $listLength]
	} else {
	    lappend ListOut [lrange $List $L2 [expr $listLength - 1]]
	}
    }
    return $ListOut
} 

#-----------------------------------------------------------------------------------------------------------------------
#------ Example of call
#if [catch {Fit_GenerateElegantFileFromLTE -eleOutputFile $eleOutputFile \
#	       -run_setup [list "lattice aps.lte use_beamline RING rootname test parameters %s.param"] \
#	       -load_parameters [list "filename_list \"$paramFileList\"" "filename sextScan.param change_defined_values 0"] \
#	       -alter_elements [list "name * type MAXAMP item X_MAX value 0" "name * type MAXAMP item Y_MAX value 0"] \
#	       -correct [list "mode trajectory n_iterations 10 n_xy_cycles 3"] \
#	       -correction_matrix_output [list "response \"%s.hrm %s.vrm\""] \
#	       -twiss_output [list "filename <TWIFILE> matched 0 reference_file <REFERENCEFILE> reference_element <REFERENCEELEMENT>"] \
#	       -run_control [list "n_steps 1"] \
#              -bunched_beam [list "n_particles_per_bunch 1"] \
#	   } result] {
#    puts stderr "Fit_GenerateElegantFileFromLTE: $result"
#    exit
#}
#------ load_parameters2 is used to add flexibility in order of load_parameters and alter_elements
#
#------ Example of call if you want to use variable in the argument (note {} around list command):
#if [catch {eval Fit_GenerateElegantFileFromLTE1 -eleOutputFile $eleFile2 \
#	       -run_setup {[list "lattice $lteFile2 use_beamline INJLINE default_order 2 centroid $tmpRoot-2.cen"]} \
#	       $loadParamString \
#	       -alter_elements {[list "name MALINJFIT item D$plane value <DZ0>" \
#				    "name MALINJFIT item D${plane}P value <DZP0>" \
#                                    "name S38IK1 item ANGLE value <IK1KICK>" \
#                                    "name S39IK2 item ANGLE value <IK2KICK>" \
#                                    "name S39IK3 item ANGLE value <IK3KICK>" \
#				     "name S40IK4 item ANGLE value <IK4KICK>"]} \
#	       -run_control {[list "n_steps 1"]} \
#	       -bunched_beam {[list "n_particles_per_bunch 1"]} \
#	   } result] {
#    return -code error "Fit_GenerateElegantFileFromLTE1: $result"
#}

proc Fit_GenerateElegantFileFromLTE1 { args } {
    #--- Order of this list defines the orbit in the ele file:
    set argList [list transmute_elements divide_elements run_setup load_parameters alter_elements load_parameters2 steering_element twiss_output0 run_control correct closed_orbit \
		     correction_matrix_output twiss_output coupled_twiss_output rf_setup moments_output optimization_setup optimization_term optimization_variable \
		     bunched_beam optimize track matrix_output]
    foreach arg $argList {set $arg ""}
    set eleOutputFile ""

    APSParseArguments [concat $argList eleOutputFile]

    array set transmute_elements_Array [list name "" type "" new_type ""]
    array set divide_elements_Array [list name "" divisions ""]
    array set run_setup_Array [list lattice "" use_beamline "" p_central_mev 7e3 default_order 1 rootname "" semaphore_file ""\
				   parameters "" centroid "" search_path ". Multipoles"]
    array set load_parameters_Array [list filename "" filename_list "" allow_missing_elements 1 allow_missing_parameters 1 \
					 change_defined_values 1 force_occurence_data 1]
    array set alter_elements_Array [list name "" type "" item "" value "" allow_missing_elements 1]
    array set run_control_Array [list n_steps 1]
    array set correct_Array [list mode orbit method global verbose 0 n_iterations 0 n_xy_cycles 0 \
				 correction_fraction [list 1.0 1.0] corrector_tweek [list 1e-5 1e-5] fixed_length 0 \
				 fixed_length_matrix 0 closed_orbit_iteration_fraction 0.5 closed_orbit_accuracy 1e-9]
    array set closed_orbit_Array [list output "" output_monitors_only 1 fixed_length 0 iteration_fraction 0.5 closed_orbit_accuracy 1e-7]
    array set correction_matrix_output_Array [list response "" output_at_each_step 1 fixed_length 0 coupled 1 response2 "" response3 ""]
    array set twiss_output_Array [list filename "" radiation_integrals 0 output_at_each_step 1 matched 1 reference_file "" \
				      reference_element ""]
    array set coupled_twiss_output_Array [list filename ""]
    array set moments_output_Array [list filename "" matched 1 output_at_each_step 1 verbosity 0 radiation 1]
    array set optimization_setup_Array [list mode minimize method simplex target 1e-14 tolerance 1e-14 n_passes 1 n_restarts 1 \
					    n_evaluations 100 log_file /dev/tty verbose 0]
    array set optimization_term_Array [list term ""]
    array set optimization_variable_Array [list name ""]
    array set bunched_beam_Array [list n_particles_per_bunch 1]
    array set track_Array [list center_on_orbit 0]
    array set optimize_Array [list summarize_setup 0]
    array set matrix_output_Array [list SDDS_output ""]
    array set rf_setup_Array [list name ""]
    array set steering_element_Array [list name ""]

    if [catch {open $eleOutputFile w} fileId] {
        return -code error "Cannot open the file $eleOutputFile for writing: $fileId"
    }
    set ifTrackCommand 0
    foreach command $argList {
	if [string length [set $command]] {
	    #------ command is a list of lists of options to allows for multiple same-name commands (like load_parameters)
	    set commandList [set $command]
	    if {[string compare $command track] == 0} {set ifTrackCommand 1}
	    if {[string compare $command optimize] == 0} {set ifTrackCommand 1}
	    if {[string compare $command load_parameters2] == 0} {set command load_parameters}
	    if {[string compare $command twiss_output0] == 0} {set command twiss_output}
	    foreach singleCommand $commandList {
		array set optionsArray $singleCommand
		array set defaultOptionsArray [array get ${command}_Array]
		foreach option [array names optionsArray] {set defaultOptionsArray($option) $optionsArray($option)}
		#------ Cannot have load_parameters with empty filename and filename_list - elegant gives error.
		if {[string compare $command load_parameters] == 0} {
		    foreach option [array names optionsArray] {set defaultOptionsArray($option) $optionsArray($option)}
		    if {![string length $defaultOptionsArray(filename)] \
			    && ![string length $defaultOptionsArray(filename_list)]} {
			unset optionsArray
			unset defaultOptionsArray
			continue
		    }
		}
		puts $fileId "&$command"
		foreach option [array names defaultOptionsArray] {
		    if [string length $defaultOptionsArray($option)] {
			if {[llength $defaultOptionsArray($option)] == 1} {
			    switch $option {
				response2 {puts $fileId "   response\[2\] = $defaultOptionsArray($option),"}
				response3 {puts $fileId "   response\[3\] = $defaultOptionsArray($option),"}
				default {puts $fileId "   $option = $defaultOptionsArray($option),"}
			    }
			} else {
			    if {[string compare $option filename_list] == 0 || [string compare $option search_path] == 0} {
				puts $fileId "   $option = \"$defaultOptionsArray($option),\""
			    } else {
				puts $fileId "   ${option}\[0\] = [join $defaultOptionsArray($option) ,],"
			    }
			}
		    }
		}
		puts $fileId "&end"
		puts $fileId " "
		unset optionsArray
		unset defaultOptionsArray
	    }
	}
    }
    if !$ifTrackCommand {
	puts $fileId "&track &end"
	puts $fileId " "
    }
    close $fileId
}

#-----------------------------------------------------------------------------------------------------------------------
# This procedure is to substitute the "echo" command

proc Fit_WriteToFile {args} {

    set usage "usage: Fit_WriteToFile -filename <filename> -accessMode <w|a> -line <string>"
    set filename ""
    set accessMode ""
    set line ""
    APSParseArguments {filename accessMode line}
    if {![string length $filename] || ![string length $accessMode] || ![string length $line]} {
        return -code error "$usage"
    }
    switch -exact -- $accessMode {
        w { 
	    if [catch {open $filename w} fid] { 
		return -code error "$fid" 
	    } 
	}
        a { 
	    if [catch {open $filename a} fid] { 
		return -code error "$fid" 
	    } 
	}
        default { return -code error "$usage" }
    }
    puts $fid $line
    close $fid
}


#-----------------------------------------------------------------------------------------------------------------------
#-----------------------------------------------------------------------------------------------------------------------
# This procedure is to pick LOCO configurations

proc Fit_ChooseLOCOConfiguration { args } {

    set configDirectory ""
    APSParseArguments {configDirectory}

    set dirList [glob -nocomplain -- $configDirectory/??? ]

    set commentList ""
    set datasetList ""
    foreach directory $dirList {
	if [file exists $directory/comment] {
	    set comment "[file tail $directory]: "
	    if [catch {open $directory/comment r} fid] {
		return -code error "$fid"
	    } else {
		lappend commentList [append comment [gets $fid]]
		close $fid
		lappend datasetList $directory
	    }
	}
    }
    set sortedCommentList [lsort -command Fit_commentCompare $commentList]
    set sortedDatasetList ""
    foreach comment $sortedCommentList {
	set i [lsearch -exact $commentList $comment]
	if {$i >= 0} {
	    lappend sortedDatasetList [lindex $datasetList $i]
	} else {
	    return -code error "Error in the comment lists!"
	}
    }
    
    set chosenList [APSChooseItemFromList \
			-name "Data Set Selection" \
			-itemList $sortedCommentList \
			-returnList $sortedDatasetList \
			-returnIndices 0 \
			-multiItem 1 \
			-contextHelp "Select saveset for data reprocessing."]

    return $chosenList
}

#----------------------------------------------------------------------------------------------------------------------------

proc Fit_commentCompare { a b } {
    set adate [string range $a 5 17]
    set bdate [string range $b 5 17]
    set res [string compare $adate $bdate]
    if {$res != 0} {
	return $res
    } else {
	return [string compare $a $b]
    }
}

#----------------------------------------------------------------------------------------------------------------------------

proc Fit_EnableButtons { widgetList } {
    foreach widget $widgetList {$widget configure -state normal}
}
proc Fit_DisableButtons { widgetList } {
    foreach widget $widgetList {$widget configure -state disabled}
}

#----------------------------------------------------------------------------------------------------------------------------
# Copies inputFile into outputFile and inserts includeString after (or instead if substitute is non-zero) searchString.

proc Fit_AddLine {args} {

    set inputFile ""
    set outputFile ""
    set searchString ""
    set includeString ""
    set substitute ""
    APSParseArguments {inputFile outputFile searchString includeString substitute}
    if {![string length $inputFile] || ![string length $outputFile] || ![string length $searchString] \
	    || ![string length $includeString] || ![string length $substitute]} {
	return -code error "Wrong arguments."
    }
    catch {exec grep $searchString $inputFile | wc} result
    set ambiguity [lindex $result 0]
    if {$ambiguity < 1} {
	return -code error "No $searchString found in $inputFile."
    }
    if {$ambiguity > 1} {
	return -code error "More than one appearance of $searchString found in $inputFile."
    }
    if [catch {open $inputFile r} fileId] {
        return -code error "Cannot open the file $inputFile for reading: $fileId"
    }
    if [catch {open $outputFile w} fileId1] {
        return -code error "Cannot open the file $outputFile for writing: $fileId1"
    }
    while {[gets $fileId line] >= 0} {
        if {[string first $searchString $line] > 0} {
	    if !$substitute { puts $fileId1 $line }
	    puts $fileId1 "$includeString "
        } else {
            puts $fileId1 $line
        }
    }
    close $fileId1
    close $fileId
}

#----------------------------------------------------------------------------------------------------------------------------

proc Fit_DeleteElementsFromList { args } {
    APSParseArguments { elementList deleteElements }
    if [llength $deleteElements] {
	foreach element $deleteElements {
	    set index [lsearch -exact $elementList $element]
	    if {$index >= 0} {
		set elementList [lreplace $elementList $index $index]
	    }
	}
    }
    return $elementList
}

#-------------------------------------------------------------------------------------------------------------------------

proc Fit_LookForDoneFiles {args} {
    APSParseArguments {jobNameList completedJobList doneFileList waitTime}
    set timesToTry 20
    for {set i 0} {$i < $timesToTry} {incr i} {
	set lostJobList ""
	foreach jobName $completedJobList {
	    if ![file exists [lindex $doneFileList [lsearch -exact $jobNameList $jobName]]] {lappend lostJobList $jobName}
	}
	if ![llength $lostJobList] { return $lostJobList }
	after $waitTime
    }
    return $lostJobList
}

#-------------------------------------------------------------------------------------------------------------------------

proc Fit_ResubmitLostJobs {args} {
    APSParseArguments {lostJobList jobNameList commandList doneFileList}
    set lostCommandList ""
    set lostDoneFileList ""
    foreach lostJob $lostJobList {
	lappend lostCommandList [lindex $commandList [lsearch -exact $jobNameList $lostJob]]
	lappend lostDoneFileList [lindex $doneFileList [lsearch -exact $jobNameList $lostJob]]
    }
    if [catch {Fit_SubmitJobs -commandList $lostCommandList -jobNameList $lostJobList \
		   -doneFileList $lostDoneFileList} returnList] {
	return -code error "SubmitJobs: $returnList"
    }
    return $returnList
}

#-------------------------------------------------------------------------------------------------------------------------
#------ Procedure for FNAL heimdall cluster:

proc Fit_GetNewJobList_PBS {args} {
    APSParseArguments {jobNameList}
    set newJobList ""
    if ![catch {join [exec qstat -f | grep Job_Name | editstring -stream "-edit=a 4D"] } fullJobList] {
	foreach jobName $fullJobList {
	    if {[lsearch -exact $jobNameList $jobName] != -1} {
		lappend newJobList $jobName
	    }
	}
    }
    return $newJobList
}

#-------------------------------------------------------------------------------------------------------------------------
#------ Procedure for APS clusters:

proc Fit_GetNewJobList_SGE {args} {
    APSParseArguments {jobNameList pidList}
    set newJobList ""
    foreach pid $pidList {
	if ![catch {join [exec qstat -j $pid | grep job_name \
			      | editstring -stream "-edit=a 2D"] } runningJobName] {
				  lappend newJobList $runningJobName
			      }
    }
    return $newJobList
}

#-------------------------------------------------------------------------------------------------------------------------
#------ Procedure for multiple tasks without queue

proc Fit_GetNewJobList_PS {args} {
    APSParseArguments {jobNameList pidList}
    if [catch {join [exec ps -p [join $pidList ,] | grep -v PID | editstring -stream "-edit=a x/ /  1F 100d"] } newPidList] {
	set newPidList ""
    }
    set newJobNameList ""
    foreach pid $newPidList {
	lappend newJobNameList [lindex $jobNameList [lsearch -exact $pidList $pid]]
    }
    return $newJobNameList
}

#-------------------------------------------------------------------------------------------------------------------------

proc Fit_WaitForTasks_New { args } {

    global useQsub queueSystemName

    set maxNumberOfResubmissions 1

    set waitTime 1800
    set usePopupWindow 0
    set waitInterval 10
    set useQsub 1
    set verbose 0
    APSParseArguments { pidList jobNameList commandList logFileList doneFileList useQsub verbose \
			    waitTime waitInterval usePopupWindow abortFile }

    switch -exact -- $queueSystemName {
	SGE { set getNewJobsProcedure Fit_GetNewJobList_SGE }
	PBS { set getNewJobsProcedure Fit_GetNewJobList_PBS }
	PS  { set getNewJobsProcedure Fit_GetNewJobList_PS }
	default {return -code error "Your Queue System Name $queueSystemName is not recognized. Valid names are SGE, PBS, PS."}
    }

    set counter 0
    set counterMax [expr $waitTime / $waitInterval]
    set waitIntervalms [format %.0f [expr $waitInterval * 1000]]
    set listLength [llength $pidList]

    after $waitIntervalms
    set oldJobList $jobNameList
    set newJobList $jobNameList
    set newPidList $pidList
    set numberOfResubmissionsList ""
    foreach element $jobNameList { lappend numberOfResubmissionsList 0 }
    while {[llength $newJobList] != 0} {
	if $verbose {puts stdout "------> New step"}
	#----- Get New Job List (the jobs that are running or waiting in queue right now)
	if [catch {$getNewJobsProcedure -jobNameList $newJobList -pidList $newPidList} newJobList] {
	    return -code error "$getNewJobsProcedure: $newJobList"
	}
	set newPidList ""
	foreach newJob $newJobList {
	    lappend newPidList [lindex $pidList [lsearch -exact $jobNameList $newJob]]
	}
	if $verbose { puts stdout "New job list: $newJobList" }
	if $verbose { puts stdout "New pid list: $newPidList" }
	#----- Get completed job list...
	if ![llength $newJobList] {
	    set completedJobList $oldJobList
	} else {
	    set completedJobList ""
	    foreach jobName $oldJobList {
		if {[lsearch -exact $newJobList $jobName] == -1} {lappend completedJobList $jobName }
	    }
	}
	if $verbose { if [llength $completedJobList] { puts stdout "Completed job list: $completedJobList" } }
	#----- Wait for files to appear...
	after $waitIntervalms
	update
	#----- Look for doneFiles...
	if [catch {Fit_LookForDoneFiles -jobNameList $jobNameList -completedJobList $completedJobList \
		       -doneFileList $doneFileList -waitTime $waitIntervalms} lostJobList] {
	    return -code error "Fit_LookForDoneFiles: $lostJobList"
	}
	if $verbose {if [llength $lostJobList] {puts stdout "Lost job list: $lostJobList"}}
	#----- Checking waiting time. If exceeds maximum, kill all left jobs and resubmit them.
	if {$counter > $counterMax} {
	    catch {eval exec qdel [join $pidList ]}
	    after 5000
	    set lostJobList $newJobList
	    set counter 0
	} else {
	    incr counter
	}
	#----- Resubmit lost jobs and jobs with errors...
	set resubmitJobList $lostJobList
	if {[llength $resubmitJobList] && $maxNumberOfResubmissions} {
	    if $verbose { puts stdout "Resubmit job list: $resubmitJobList" }
	    if [catch {Fit_ResubmitLostJobs -lostJobList $lostJobList -jobNameList $jobNameList -commandList $commandList \
			   -doneFileList $doneFileList} resubmittedList] {
		return -code error "Fit_ResubmitLostJobs: $resubmittedList"
	    }
	    set resubmittedPidList [lindex $resubmittedList 0]
	    set resubmittedLogFileList [lindex $resubmittedList 1]
	    foreach jobName $resubmitJobList pidNumber $resubmittedPidList logFile $resubmittedLogFileList {
		set I [lsearch -exact $jobNameList $jobName]
		set pidList [lreplace $pidList $I $I $pidNumber]
		set logFileList [lreplace $logFileList $I $I $logFile]
		set numberOfResubmissionsList \
		    [lreplace $numberOfResubmissionsList $I $I [expr [lindex $numberOfResubmissionsList $I] + 1]]
	    }
	    foreach element $numberOfResubmissionsList jobName $jobNameList {
		if {$element > $maxNumberOfResubmissions} {
		    if $useQsub {
			catch {eval exec qdel [join $pidList ]}
		    } else {
			catch {eval exec kill -9 [join $pidList ]}
		    }
		    return -code error "Error: Job $jobName required resubmission $element times. Exiting and killing all remaining jobs."
		}
	    }

	}
	#----- CheckAbortStatus
	if [file exists $abortFile] {
	    if $useQsub {
		catch {eval exec qdel [join $pidList ]}
	    } else {
		catch {eval exec kill -9 [join $pidList ]}
	    }
	    return -code error "Interrupted by user."
	}
    }
    catch {eval file delete $logFileList}
}

#-------------------------------------------------------------------------------------------------------------------------

proc Fit_SubmitJobs {args} {
    global env qsubCommand useQsub qsubRespProcCommand submissionPause
    set verbose 0
    APSParseArguments { commandList jobNameList doneFileList verbose }
    set pidList ""
    set logFileList ""
    set scriptFileList ""
    foreach command $commandList jobName $jobNameList doneFile $doneFileList {
	set path [file dirname $doneFile]
	set scriptFile $path/$jobName.tcl
	if [catch {open $scriptFile w} fid] {
	    return -code error "Error opening file: $fid"
	}
	puts $fid "#!/bin/sh  "
	puts $fid " "
	puts $fid "$command"
	close $fid
	exec chmod +x $scriptFile
	catch {file delete $path/$jobName.stderr}

	if $verbose { puts stdout "submitCommand: $qsubCommand" }
	if [catch {eval $qsubCommand} qsubResponse] {
	    return -code error "Error submitting command: $qsubResponse"
	}
	if $verbose { puts stdout "submit response: $qsubResponse" }
	set pid [eval $qsubRespProcCommand]
	if $verbose { puts stdout "Job PID is $pid" }
	lappend pidList $pid
	lappend logFileList $path/$jobName.stderr
	lappend scriptFileList $scriptFile
	lappend dotOFileList $scriptFile.o$pid
	#--- Pause for submissionPause seconds (after takes milliseconds).
	if $submissionPause { after [expr $submissionPause * 1000] }
    }
    return [list $pidList $logFileList $scriptFileList $dotOFileList]
}

#-------------------------------------------------------------------------------------------------------------------------

proc Fit_CalculateResponseMatrixDerivative { args } {

    global qsubCommand useQsub submissionPause qsubRespProcCommand queueSystemName
    set verbose 0
    set deleteFilesOnExit 1
    set argCopy $args
    APSParseArguments {varList splitTasks scriptName scriptParameters matrixFile tmpDirName useQsub qsubCommand \
				 continuePrevious rootTaskName usePopupWindow waitTime waitInterval deleteFilesOnExit \
				 submissionPause abortFile verbose qsubRespProcCommand queueSystemName}
    if $verbose {puts stdout "Fit_CalculateResponseMatrixDerivative agruments: $argCopy"}

    if [catch {Fit_SplitList -List $varList -splitTasks $splitTasks} listAfterSplitting] {
	return -code error "Fit_SplitList: $listAfterSplitting"
    }
    
    if ![string length $rootTaskName] {
	set rootTaskName [string range [file rootname [file tail $scriptName]] 0 7]
    }

    set totalPathList ""
    set commandList ""
    set jobNameList ""
    set doneFileList ""
    for {set jobNumber 0} {$jobNumber < [llength $listAfterSplitting]} {incr jobNumber} {
	set elements [lindex $listAfterSplitting $jobNumber]
	set totalPath $tmpDirName/$jobNumber
	set jobName $rootTaskName$jobNumber
	lappend totalPathList $totalPath

	if $continuePrevious {
	    if [file exists $totalPath/task.completed] {
		continue
	    }
	}

	lappend jobNameList $jobName
	lappend doneFileList $totalPath/task.completed
	catch {exec mkdir $totalPath}
	catch {eval file delete -force [glob -nocomplain -- $totalPath/*]}

	set scriptParametersFile $totalPath/scriptParameters
	if [catch {open $scriptParametersFile w} fid] { 
	    return -code error "$fid" 
	}
	puts $fid "-path $totalPath -varList \"$elements\" $scriptParameters"
	close $fid

	set commandLine "$scriptName -argMode file -scriptParametersFile $scriptParametersFile -verbose $verbose -deleteFilesOnExit $deleteFilesOnExit"
	lappend commandList $commandLine
    }

    if [catch {Fit_SubmitJobs -commandList $commandList -jobNameList $jobNameList \
		   -doneFileList $doneFileList -verbose $verbose} returnList] {
	return -code error "Fit_SubmitJobs: $returnList"
    }
    set pidList [lindex $returnList 0]
    set logFileList [lindex $returnList 1]
    if $useQsub {
	if [catch {Fit_WaitForTasks_New -pidList $pidList -jobNameList $jobNameList -commandList $commandList \
		       -logFileList $logFileList -doneFileList $doneFileList -waitTime $waitTime \
		       -waitInterval $waitInterval -usePopupWindow $usePopupWindow -verbose $verbose \
		       -useQsub $useQsub -abortFile $abortFile} result] {
	    return -code error "Fit_WaitForTasks_New: $result"
	}
    }

    set listLength [llength $totalPathList]
    set matrixFileList ""
    foreach path $totalPathList {
        lappend matrixFileList $path/responseMatrix.sdds
    }
    if {[llength $matrixFileList] == 1} {
        file copy -force $matrixFileList $matrixFile
    } else {
	if [catch {eval exec nice -19 sddsxref $matrixFileList $matrixFile "-take=* -leave=Rootname"} result] {
	    return -code error "Error during final combining: (list of xrefed files is: $matrixFileList) $result"
	} else {
	    if $deleteFilesOnExit {
		for {set I 0} {$I < $listLength} {incr I} {
		    set path [lindex $totalPathList $I]
		    catch {eval file delete [glob -nocomplain -- $path/*]}
		    catch {file delete $path}
		}
	    }
	}
    }
    return 0
}

#-----------------------------------------------------------------------------------------------------------------------

proc Fit_ReadListFromFile { args } {
    set optionsList ""
    APSParseArguments { listFile optionsList }
    set mainList ""
    if [string length $optionsList] {
	array set options $optionsList
	set indexList [list useQuads useCorrs useCorrs useBPMs useBPMs useEnergy useSkew useCorrsTilt useCorrsTilt \
	    useBPMsTilt useBPMsTilt]
	set paramList [list Quad HCorr VCorr HBPM VBPM Energy Skew HCorrTilt VCorrTilt HBPMTilt VBPMTilt]
	foreach index $indexList param $paramList {
	    if $options($index) {
		set localList [join [exec sddsprocess $listFile -pipe=out -match=para,ListName=$param -nowarning \
					 | sdds2stream -pipe=in -col=List] ]
		lappend mainList $localList
	    } else {
		lappend mainList ""
	    }
	}
    } else {
	foreach param [list HCorr VCorr HBPM VBPM] {
	    lappend mainList [join [exec sddsprocess $listFile -pipe=out -match=para,ListName=$param -nowarning \
					| sdds2stream -pipe=in -col=List] ]
	}
    }
    return $mainList
}

#-------------------------------------------------------------------------------------------------------------------------
#------ Xbpm = ( Xorb * Cos A + Yorb * Sin A) * (1 + Gx)
#------ Ybpm = (-Xorb * Sin A + Yorb * Cos A) * (1 + Gy)
#------ where Zbpm  -  bpm reading
#------       Zorb  -  real orbit
#------          A  -  tilt
#------         Gz  -  bpm gain error
#---
proc Fit_ApplyBPMCorrections { args } {
    global tmpDir simplifiedTiltTreatment
    set bpmNoise 0
    set seed 0
    APSParseArguments {hrmFile hrmXFile vrmFile vrmXFile coupledMatrix bpmGainXFile bpmGainYFile bpmTiltXFile bpmTiltYFile bpmNoise seed}
    
    set tmpRoot $tmpDir/[APSTmpString]-bpmCorrect
    set deleteFiles ""

    set xbpmTiltFile $bpmTiltXFile
    set ybpmTiltFile $bpmTiltYFile
    set xbpmGainFile $bpmGainXFile
    set ybpmGainFile $bpmGainYFile

    if $simplifiedTiltTreatment {
	set doTiltCorrection 0
	if [file exists $xbpmTiltFile] {
	    if {[exec sdds2stream $xbpmTiltFile -rows=bare]} {set doTiltCorrection 1}
	}
    } else {
	set doTiltCorrection 1
	if {![file exists $xbpmTiltFile] || ![file exists $ybpmTiltFile] || ![file exists $hrmXFile] || ![file exists $vrmXFile]} {set doTiltCorrection 0}
    }
    set doGainCorrection 0
    if {[file exists $bpmGainXFile] && [file exists $bpmGainYFile]} {
	if {[exec sdds2stream $bpmGainXFile -rows=bare] && [exec sdds2stream $bpmGainYFile -rows=bare]} {set doGainCorrection 1}
    }

    #------ Tilt correction
    #------ Simplified tilt treatment is when I only use xbpmTiltFile for unique angle definition (no xAngle/yAngle)
    if $doTiltCorrection {
	if $simplifiedTiltTreatment {
	    foreach rmFile [list $hrmFile $vrmFile] Plane [list X Y] {
		exec sddsprocess $xbpmTiltFile $tmpRoot.tilts "-reedit=col,TagName,%@#X@#$Plane@"
		foreach plane [list x y] pageIndex [list 1 2] {
		    foreach trigFunc [list cos sin] {
			exec sddsconvert $rmFile -pipe=out -keep=$pageIndex -rename=col,BPMName=BPMNameExclude,TagName=TagNameExclude \
			    | sddsxref -pipe $tmpRoot.tilts -take=ParameterValue -match=TagNameExclude=TagName -fill -nowarning \
			    | sddsprocess -pipe "-redef=col,%s,%s ParameterValue $trigFunc *,select=*,exclude=*Exclude" \
			    | sddsconvert -pipe=in $tmpRoot.$plane.$trigFunc -rename=col,BPMNameExclude=BPMName,TagNameExclude=TagName -del=col,ParameterValue
		    }
		}
		exec sddsmatrixop $tmpRoot.x.cos -pipe=out -push=$tmpRoot.y.sin -add \
		    | sddsxref -pipe=in $tmpRoot.x.cos $tmpRoot.x -take=BPMName,TagName -transfer=para,*
		exec sddsmatrixop $tmpRoot.x.sin -pipe=out -scalarmult=-1 -push=$tmpRoot.y.cos -add \
		    | sddsxref -pipe=in $tmpRoot.y.cos $tmpRoot.y -take=BPMName,TagName -transfer=para,*
		exec sddscombine $tmpRoot.x $tmpRoot.y $rmFile -overwrite
		lappend deleteFiles $tmpRoot.x.cos $tmpRoot.x.sin $tmpRoot.y.cos $tmpRoot.y.sin $tmpRoot.x $tmpRoot.y $tmpRoot.tilts
	    }
	} else {
	    set xbpms [lindex [exec sdds2stream $hrmFile -rows=bare] 0]

	    exec sddsconvert $hrmFile $tmpRoot.xbpms -keep=1 -retain=col,TagName
	    exec sddsconvert $hrmFile $tmpRoot.ybpms -keep=2 -retain=col,TagName
	    exec sddsxref $tmpRoot.xbpms $xbpmTiltFile $xbpmTiltFile.full -take=ParameterValue -match=TagName -fill -nowarning
	    exec sddsxref $tmpRoot.ybpms $ybpmTiltFile $ybpmTiltFile.full -take=ParameterValue -match=TagName -fill -nowarning
	    lappend deleteFiles $tmpRoot.xbpms $tmpRoot.ybpms $xbpmTiltFile.full $ybpmTiltFile.full

	    exec sddsprocess $xbpmTiltFile.full $tmpRoot.sin1 "-redef=col,SinValue,ParameterValue sin"
	    exec sddsprocess $ybpmTiltFile.full $tmpRoot.sin2 "-redef=col,SinValue,ParameterValue sin -1 *"
	    exec sddscombine $xbpmTiltFile.full $ybpmTiltFile.full -pipe=out -merge \
		| sddsprocess -pipe=in $tmpRoot.directVector "-redef=col,CosValue,ParameterValue cos"
	    exec sddscombine $tmpRoot.sin1 $tmpRoot.sin2 $tmpRoot.crossVector -merge -overWrite
	    lappend deleteFiles $tmpRoot.sin1 $tmpRoot.sin2 $tmpRoot.directVector $tmpRoot.crossVector

	    foreach rmFile [list $hrmFile $vrmFile $hrmXFile $vrmXFile] \
		vectorFile [list $tmpRoot.directVector $tmpRoot.directVector $tmpRoot.crossVector $tmpRoot.crossVector] \
		columnName [list CosValue CosValue SinValue SinValue] \
		extName [list hDirect vDirect hCross vCross] {
		    if [catch {exec sddscombine $rmFile -pipe=out -merge \
				   | sddsxref -pipe $vectorFile -nowarning -fillIn -take=$columnName -match=TagName \
				   | sddsconvert -pipe -rename=col,BPMName=BPMNameValue,TagName=TagNameValue \
				   | sddsprocess -pipe "-redef=col,%s,%s $columnName *,select=*,exclude=*Value" \
				   | sddsconvert -pipe=in $tmpRoot.$extName -rename=col,BPMNameValue=BPMName,TagNameValue=TagName \
				   -del=col,$columnName} result] {
			return -code error "Error applying bpm tilts (rmFile $rmFile; vectorFile $vectorFile): $result"
		    }
		    lappend deleteFiles $tmpRoot.$extName
		}
	    exec sddsmatrixop $tmpRoot.hDirect -pipe=out -push=$tmpRoot.hCross -add \
		| sddsbreak -pipe -rowlimit=$xbpms \
		| sddsxref -pipe=in $hrmFile $tmpRoot.hResult -take=BPMName,TagName -transfer=para,*
	    exec sddsmatrixop $tmpRoot.vDirect -pipe=out -push=$tmpRoot.vCross -add \
		| sddsbreak -pipe -rowlimit=$xbpms \
		| sddsxref -pipe=in $vrmFile $tmpRoot.vResult -take=BPMName,TagName -transfer=para,*
	    lappend deleteFiles $tmpRoot.hResult $tmpRoot.vResult
	    file copy -force $tmpRoot.hResult $hrmFile
	    file copy -force $tmpRoot.vResult $vrmFile
	}
    }

    #------ Gain correction
    if $doGainCorrection {
	set rmFileList [list $hrmFile $vrmFile]
	set planeList [list X Y]
	set bpmFile $tmpRoot.bpmGain
	foreach rmFile $rmFileList plane $planeList {
	    switch -regexp -- $coupledMatrix {
		0|1 { 
		    switch -exact -- $plane {
			X {set bpmFile $xbpmGainFile}
			Y {set bpmFile $ybpmGainFile}
		    }
		}
		2 {exec sddscombine $xbpmGainFile $ybpmGainFile $bpmFile -overWrite; lappend deleteFiles $bpmFile}
	    }
	    if [exec sddscombine $bpmFile -pipe=out -merge | sdds2stream -pipe=in -rows=bare] {
		if [catch {exec sddsconvert $rmFile -pipe=out -rename=col,BPMName=BPMNameValue,TagName=TagNameValue \
			       | sddsxref -nowarning -pipe $bpmFile -fillIn -take=ParameterValue -match=TagNameValue=TagName \
			       | sddsprocess -pipe "-redef=col,%s,%s 1 ParameterValue + *,select=*,exclude=*Value" \
			       | sddsconvert -pipe=in $tmpRoot.gainResult -del=col,ParameterValue \
			       -rename=col,BPMNameValue=BPMName,TagNameValue=TagName} result] {
		    return -code error "BPM gain correction error: $result"
		}
		file copy -force $tmpRoot.gainResult $rmFile
		lappend deleteFiles $tmpRoot.gainResult 
	    }
	}
    }

    #------ Adding bpm noise (dispersion is without noise for now):
    if {$bpmNoise > 0} {
	foreach rmFile [list $hrmFile $vrmFile] doDisp [list 1 0] {
	    incr seed 5000
	    if $doDisp {set dispString " -redef=col,Dispersion,0 "} else {set dispString ""}
	    if [catch {eval exec sddsprocess $rmFile -pipe=out \"-rpnexpression=$seed srnd\" \
			   \"-redef=col,%s,3 grndl $bpmNoise *,select=*,exclude=*Name\" $dispString \
			   | sddsmatrixop -pipe -push=$rmFile -add \
			   | sddsxref -pipe=in $rmFile $tmpRoot.bpmNoise -take=BPMName,TagName -transfer=para,*} result] {
		return -code error "Error adding BPM noise ($rmFile): $result"
	    }
	    file copy -force $tmpRoot.bpmNoise $rmFile
	}
	lappend deleteFiles $tmpRoot.bpmNoise
    }
    catch {eval file delete $deleteFiles}
}

#-------------------------------------------------------------------------------------------------------------------------
#--- Still not completed, but should be...
proc ApplyBPMCorrectionNew {args} {
    global tmpDir options
    APSParseArguments {hrmFile vrmFile xGainFile yGainFile tiltFile coupledMatrix residualError}
###    APSParseArguments {rmRootname bpmErrorsFile rmKick noiseSigma}
    set tmpRoot /tmp/[APSTmpString]-bpmErrors

    exec sddsprocess $hrmFile -pipe=out -match=para,Plane,X | sddsconvert -pipe=in $tmpRoot.hxrm -del=col,s
    exec sddsprocess $hrmFile -pipe=out -match=para,Plane,Y | sddsconvert -pipe=in $tmpRoot.hyrm -del=col,s
    exec sddsprocess $vrmFile -pipe=out -match=para,Plane,X | sddsconvert -pipe=in $tmpRoot.vxrm -del=col,s
    exec sddsprocess $vrmFile -pipe=out -match=para,Plane,Y | sddsconvert -pipe=in $tmpRoot.vyrm -del=col,s
    lappend deleteFiles $tmpRoot.hxrm $tmpRoot.hyrm $tmpRoot.vxrm $tmpRoot.vyrm

    #--- Build cosine, sin, gainX, gainY files that have the same dimension as rm files to use in hadamard multiplication:
    #--- (gainx_BPM1 gainx_BPM1 gainx_BPM1 ...)
    #--- (gainx_BPM2 gainx_BPM2 gainx_BPM2 ...)
    #--- (gainx_BPM3 gainx_BPM3 gainx_BPM3 ...)
    #--- ( ...                                )
    #--- hrm and vrm files are treated separately because they could have different number of correctors
    #--- x and y bpm number must be the same (filtering should be done later)
    foreach plane [list h v] {
	foreach CosSin [list cos sin] {
	    exec sddsxref $tmpRoot.${plane}xrm $tiltFile -pipe=out -take=Tilt \
		| sddsconvert -pipe -del=col,BPMName \
		| sddsprocess -pipe "-redef=col,%s,Tilt $CosSin,select=*,exclude=Tilt" \
		| sddsconvert -pipe=in $tmpRoot.$plane$CosSin -del=col,Tilt
	    lappend deleteFiles $tmpRoot.$plane$CosSin
	}
	foreach col [list GainX GainY] {
	    exec sddsxref $tmpRoot.${plane}xrm $bpmErrorsFile -pipe=out -take=$col \
		| sddsconvert -pipe -del=col,BPMName \
		| sddsprocess -pipe "-redef=col,%s,1 $col +,select=*,exclude=$col" \
		| sddsconvert -pipe=in $tmpRoot.$plane$col -del=col,$col
	    lappend deleteFiles $tmpRoot.$plane$col
	}
    }

    #--- Use formula above directly:
    foreach xOrbFile [list $tmpRoot.hrm $tmpRoot.hvrm] yOrbFile [list $tmpRoot.vhrm $tmpRoot.vrm] \
	newxFile [list $tmpRoot-new.hrm $tmpRoot-new.hvrm] newyFile [list $tmpRoot-new.vhrm $tmpRoot-new.vrm] \
	cosFile [list $tmpRoot.hcos $tmpRoot.vcos] sinFile [list $tmpRoot.hsin $tmpRoot.vsin] \
	gainxFile [list $tmpRoot.hGainX $tmpRoot.vGainX] gainyFile [list $tmpRoot.hGainY $tmpRoot.vGainY] {
	    exec sddsmatrixop $xOrbFile -pipe=out                -push=$cosFile -mult=had -push=$yOrbFile -push=$sinFile -mult=had -add -push=$gainxFile -mult=had \
		| sddsprocess -pipe "-redef=col,%s,%s,select=*,exclude=BPMName,units=m/rad" \
		| sddsxref -pipe=in $xOrbFile $newxFile -take=BPMName
	    exec sddsmatrixop $xOrbFile -pipe=out -scalarmult=-1 -push=$sinFile -mult=had -push=$yOrbFile -push=$cosFile -mult=had -add -push=$gainyFile -mult=had \
		| sddsprocess -pipe "-redef=col,%s,%s,select=*,exclude=BPMName,units=m/rad" \
		| sddsxref -pipe=in $yOrbFile $newyFile -take=BPMName
	    lappend deleteFiles $newxFile $newyFile 
	}
    
    #--- Add random noise (multiply by $rmKick since response matrix is in m/rad):
    set seed [exec rpnl "rnd 1000 * int"]
    lappend deleteFiles $tmpRoot.noise
    foreach ext [list hrm vrm hvrm vhrm] {
	exec sddsprocess $tmpRoot.$ext $tmpRoot.noise "-rpnexpression=$seed srnd" "-redef=col,%s,3 grndl $noiseSigma * $rmKick /,select=*,exclude=BPMName"
	exec sddsmatrixop $tmpRoot-new.$ext -pipe=out -push=$tmpRoot.noise -add \
	    | sddsxref -pipe=in $rmRootname.$ext $rmRootname-new.$ext -take=s,BPMName
	incr seed 5
    }
    eval file delete $deleteFiles 
}
#-------------------------------------------------------------------------------------------------------------------------
